import graph
import sys,os,copy
import networkx as nx
import xml.etree.ElementTree as ET

# Interface to use graph.py on TUNA style XML trial files

relations = ["below","on-top-of","in-front-of","behind","left-of","right-of"]

# reads in a TUNA XML trial file and converts its DOMAIN element into a scene graph, which is returned. 
def XMLtrial_to_graph(XMLtrial,Costs):	
	targetid = ""
	G = nx.MultiDiGraph()
	trial = ET.fromstring(XMLtrial)
	for entity in trial.find("DOMAIN"):
		node_id = entity.attrib["ID"]
		G.add_node(node_id)
		for attribute in entity:
			if not attribute.attrib["NAME"] == "rel_location":
				G.add_edge(node_id,node_id,att=attribute.attrib["NAME"],val=attribute.attrib["VALUE"],cost=Costs[attribute.attrib["NAME"]][attribute.attrib["VALUE"]])
			else:
				G.add_edge(node_id,attribute.attrib["REL"],att=attribute.attrib["NAME"],val=attribute.attrib["VALUE"],cost=Costs[attribute.attrib["NAME"]][attribute.attrib["VALUE"]])
		if entity.attrib["TYPE"] == "target":
			targetid = node_id
	return G,targetid
	
def get_human_desc(XMLtrial):
	trial = ET.fromstring(XMLtrial)
	return ET.tostring(trial.find("ATTRIBUTE-SET"))
	

def desc_to_XML(desc):
	XML_desc = ET.Element("ATTRIBUTE-SET")
	for att in desc:
		if not att[1] in relations:
			XML_att = ET.SubElement(XML_desc,"ATTRIBUTE")
			XML_att.attrib["ENT"] = att[0]
			XML_att.attrib["NAME"] = att[1]
			XML_att.attrib["VALUE"] = att[2]
		else:
			XML_att = ET.SubElement(XML_desc,"ATTRIBUTE")
			XML_att.attrib["ENT"] = att[0]
			XML_att.attrib["NAME"] = "rel_location"
			XML_att.attrib["VALUE"] = att[1]
			XML_att.attrib["REL"] = att[2]		
	return ET.tostring(XML_desc)
	
def make_dice_output(human_desc,XML_RE,trial_id):	
	return """<DOCUMENT>
<TRIAL ID="%s">
<INSTANCE TYPE="human">
%s
</INSTANCE>
<INSTANCE TYPE="system">
%s
</INSTANCE>
</TRIAL>
</DOCUMENT>""" %(trial_id,human_desc,XML_RE)

def main(input_para,prefcostpath,outfolder,force_type,maxlen):

	if not os.path.exists(input_para):
		print "\nERROR! Input file or folder not found! "+input_para+"\n"
		sys.exit()
	if not os.path.exists(outfolder):
		os.makedirs(outfolder)
	
	prefcostfile = open(prefcostpath,"r")
	
	PO = []
	Costs = {}
	for line in prefcostfile.readlines():
		if line != "\n":
			listed = line.split()
			if not listed[0] in Costs:
				Costs[listed[0]] = {}
			Costs[listed[0]][listed[1]] = float(listed[2])
			PO.append(listed[0]+"."+listed[1])	 

	triallist = []
	if os.path.isfile(input_para):
		triallist.append(input_para)	
	elif os.path.isdir(input_para):
		for file in os.listdir(input_para):
			triallist.append(os.path.join(input_para,file))	

	for file in triallist:
		if not ".DS_" in file:
#			print file

			H = nx.MultiDiGraph()
	
			trial_file = open(file,"r")
			trial_string = trial_file.read()
			trial_file.close()
			G,targetid = XMLtrial_to_graph(trial_string,Costs) # the domain (aka scene) graph and the targetid

			H.add_node(targetid) # the initial description graph containing only the target node

			human_desc = get_human_desc(trial_string)[:-1]
				
			bestG = None
			
			edges_to_try = graph.order_edges(graph.get_new_neighbour_edges(G,H,targetid),PO)
					
			RE = graph.find_RE_Graph(targetid,bestG,H,G,PO,edges_to_try,maxlen)
			
	#		ALWAYS ADD TYPE?
			if force_type == "all" or force_type == "a":
				graph.add_type_for_all(RE,G)
			elif force_type == "target" or force_type =="t":
				graph.add_target_type(RE,G,targetid)
			elif force_type == "none" or force_type =="n":
				pass
			else:
				print '\nPlease use "all", "none", OR "target" to specify whether type should be force-added for all objects, no objects, or only the target. (the last paramter)\n'
				sys.exit()
			
			string_RE = graph.make_short_RE(RE.edges(data=True)) 
	
			XML_RE = desc_to_XML(string_RE)
			
			infilename = os.path.basename(file)
			trial_id = infilename[:infilename.find(".")]
			DICE_output = make_dice_output(human_desc,XML_RE,trial_id)
			outfile = open(os.path.join(outfolder,"_"+infilename),"w")
			outfile.write(DICE_output)
			outfile.close()
			print string_RE


if __name__ == "__main__":
	if len(sys.argv) != 6:
		print "\nWrong number of arguments!\n"
		print "Usage: python",sys.argv[0],"input prefcostfile outfolder force-type maxlen\n"
		print "Example: python",sys.argv[0],'../input/GRE3D7-2 tg ../PrefAndCosts.txt ../data/output/ yes 4 OR:'
		print "Example: python",sys.argv[0],'../input/ tg ../PrefAndCosts.txt ../Costs.txt ../data/output/ yes 4\n'
		print '"input" can be a single scenefile or a directory of scenefiles.\n'
		print 'With the "force-type" parameter you can specify whether the "type" attribute should always be added for all objects mentioned in the description or only for the target object or not at all. Use "all", "none" or "target" (or "a", "n", "t").\n'
		sys.exit()

	input_para = sys.argv[1] # an XML file in TUNA format, containing a trial and a human-produced description or a folder containing such XML files
#	targetid = sys.argv[2] # a string specifying the id of one of the objects in trial file, usually "tg"
	prefcostpath = sys.argv[2] # a txt file listing attribute value pairs in the order of preference and with their associated costs
	outfolder = sys.argv[3]
	force_type = sys.argv[4] # "yes" or "no"
	maxlen = sys.argv[5] # an integer
	
	main(input_para,prefcostpath,outfolder,force_type,maxlen)
	
